from enum import auto

from strenum import UppercaseStrEnum


class Isotope(UppercaseStrEnum):
    H1 = auto()
    H2 = auto()
    H3 = auto()
    N15 = auto()
    C13 = auto()
    O17 = auto()
    F19 = auto()
    P31 = auto()


# fmt: off
GAMMA_RATIOS = {
    # Nuc	        ratio	        # (MHz⋅T−1)
    Isotope.H1:     1.0,            # 42.58
    Isotope.H2:     0.153508386,    # 6.54
    Isotope.H3:     1.066643718,    # 45.42
    Isotope.C13:    0.251503855,    # 10.71
    Isotope.N15:    0.101368145,    # -4.32 - should really be negative !
    Isotope.O17:    0.135564627,    # -5.77 - should really be negative !
    Isotope.F19:    0.94129576,     # 40.08
    Isotope.P31:    0.404791467,    # 17.24
}

ATOM_TO_ISOTOPE = {
    # Nuc   Isotope
    "H": Isotope.H1,
    "D": Isotope.H2,
    "T": Isotope.H3,
    "C": Isotope.C13,
    "N": Isotope.N15,
    "O": Isotope.O17,
    "F": Isotope.F19,
    "P": Isotope.P31,
}
# fmt: on



import string
from collections import namedtuple
from typing import List, Tuple

from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.sequence_lib import (
    TRANSLATIONS_1_3,
    MoleculeType,
    residues_to_residue_name_lookup,
)
from nef_pipelines.lib.structures import AtomLabel, LineInfo, Residue, SequenceResidue
from nef_pipelines.lib.util import exit_error, strip_characters_left

#
# sparky docs
#
# Make peaks on a spectrum from a peak list file (rp).
#
# Peaks are read from a peak list file and placed on the selected spectrum.
# The file should contain a line for each peak. The line should have an
# assignment followed by chemical shifts for each axis. For example
#
# 	C2H5-G1H1      5.395      6.030
# The assignment can contain ? components. It can omit a group name for a component
# -- the shorthand G1H1'-H2' where the second group is omitted is equivalent to G1H1'-G1H2'.
# The residue name is separated from the atom name by looking for a residue number followed
# by one of the letters H, C, N, Q, or M (upper or lower case). Extra columns become the
# peak note. Peaks for 3-D or 4-D spectra can be read.
#
# what does sparky do with multiple chains?
#
# what does ga mean in the sparky output after a volume
#
# why does sparky sometime have ga after a volume I presume it's a gausiann peak fit?
# comment gst 19/04/2023 is ga for a gaussian peak fit?
#
#  Assignment       w1      w2      Volume     lw1 (hz)   lw2 (hz)
#
#    G16H3'-H8    4.905   8.010   7.15e+06 ga    28.6       20.0
#    G16H4'-H8    4.439   8.013   5.42e+06 ga    35.3       16.9
#  T17H6-G16H8    7.205   8.004   1.68e+06
#  T17H7-G16H8    1.459   8.008   2.09e+07 ga    27.5       24.1
#   T17H2"-H1'    2.509   5.840   4.68e+07 ga    41.2       17.6
#
# what does ta display in a sparky project
# how does sparky cope with complexes where peaks are between chains
# which it doesn't define
#

SparkyAtomLabel = namedtuple("SparkyAtomLabel", "residue_number residue_name atom_name")


def parse_single_assignment(
    assignment: str,
    previous_residue_name: str = None,
    previous_residue_number: str = None,
) -> Tuple[str, str, str]:

    """
    Parse a single sparky assignment of the form ?, G16H4', H8, T17H6, G16H8 into a tuple of residue_number,
    residue_name and atom_name. To allow for processing of abbreviated assignments of the form G16H4'-H8 a previous
    residue number and residue name may be passed in [in this casefor H8  they would be G and 16 to match G16H4']

    :param assignment: the single sparky assignments
    :param previous_residue_name: a previous residue name to use for an abbreviated assignment
    :param previous_residue_number: a previous residue number to use for an abbreviated assignment
    :return: a tuple of the form (residue_number, residue_name , atom_name)
    """

    target = assignment

    # remove a residue name or question mark
    residue_name, target = strip_characters_left(target, string.ascii_letters)
    if len(target) == 0:
        target = residue_name
        residue_name = ""

    if target[0] == "?" and len(target) > 0:
        target = target[1:]
        residue_name = ""

    # remove residue numbers or question mark
    residue_number, target = strip_characters_left(target, string.digits)

    if len(residue_number) == 0 and len(target) != 0 and target[0] == "?":
        target = target[1:]
        residue_number = ""

    if len(target) > 0:
        atom_name = target
    else:
        atom_name = ""

    # residue name is separated from the atom name by looking for a residue number followed
    # by one of the letters H, C, N, Q, or M ergo a residue name that starts with one of these
    # must be wrong, and we have split an atom (though we don't actually need this test?)
    if residue_name != "" and residue_name.lower() in "HCNQM".lower():
        atom_name = f"{residue_name}{residue_number}{atom_name}"
        residue_name = ""
        residue_number = ""

    if (
        residue_number == ""
        and previous_residue_number is not None
        and residue_name == ""
        and previous_residue_name is not None
        and atom_name != ""
    ):
        residue_number = previous_residue_number
        residue_name = previous_residue_name

    if residue_name == "" and residue_number == "":
        for i, character in enumerate(atom_name):
            if character.lower() in "HCNQM".lower():
                if i > 0 and atom_name[i - 1] in string.digits:
                    residue_name_and_number = atom_name[:i]
                    atom_name = atom_name[i:]

                    residue_name = residue_name_and_number.rstrip(string.digits)

                    residue_number = residue_name_and_number[-len(residue_name) :]

    return SparkyAtomLabel(residue_number, residue_name, atom_name)


def parse_assignments(
    assignments: str,
    chain_code: str,
    sequence: List[SequenceResidue],
    molecule_type: MoleculeType,
    line_info: LineInfo,
) -> List[AtomLabel]:

    """
    take a sparky assignment of the form ?, G16H4'-H8 or T17H6-G16H8 into atom labels

    :param assignments: a single assignment or a - separated assignment [including abbreviated assignments]
    :param chain_code: the chain code to use
    :param sequence: the sequence to get residue names from and to check the read residue names agains
    :param molecule_type: the type of the molecule DNA RNS protein etc.
    :param line_info: file and line  information for error reporting
    :return: a list of AtomLabels
    """

    residue_name_lookup = residues_to_residue_name_lookup(sequence)

    residue_name_translations = TRANSLATIONS_1_3[molecule_type]

    fields = assignments.split("-")

    result = []
    last_sequence_code = None
    last_residue_name = None

    for field in fields:

        sequence_code, residue_name, atom_name = parse_single_assignment(
            field,
            previous_residue_number=last_sequence_code,
            previous_residue_name=last_residue_name,
        )

        if sequence_code == "":
            sequence_code = UNUSED
        if residue_name == "":
            residue_name = UNUSED

        if residue_name_translations is not None and residue_name is not UNUSED:

            if residue_name in residue_name_translations:
                translated_residue_name = residue_name_translations[residue_name]
            else:
                msg = f"""
                    The residue name {residue_name} is not defined for the molecule type {molecule_type}
                    at line {line_info.line_no} in file {line_info.file_name} the line was
                    {line_info.line}
                """
                exit_error(msg)

        residue_name_key = (chain_code, sequence_code)

        if residue_name == UNUSED and residue_name_lookup is None:
            msg = f"""
                no sequence was loaded and the assignment {assignments} while it has a sequence code {sequence_code}
                and chain code {chain_code} doesn't define a residue type, please provide a sequence in the input
                stream
            """
            exit_error(msg)
        else:

            if (
                residue_name_lookup is not None
                and residue_name_key in residue_name_lookup
            ):
                translated_residue_name = residue_name_lookup[residue_name_key]

        if len(sequence) > 0 and (residue_name_key not in residue_name_lookup):
            msg = f"""
                the chain code {chain_code} and sequence_code {sequence_code} from
                line {line_info.line_no} in file {line_info.file_name} were not found
                in the input sequence, the full line was

                {line_info.line}

                if you wish to input the peaks without validating against the input sequence use the
                --no-validate option of sparky import peaks
            """
            exit_error(msg)

        residue = Residue(
            chain_code=chain_code,
            sequence_code=sequence_code,
            residue_name=translated_residue_name,
        )
        assignment = AtomLabel(residue, atom_name)

        result.append(assignment)

        last_sequence_code = sequence_code
        last_residue_name = residue_name

    return result



import os
from textwrap import dedent

import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrview.exporters.sequences import sequences

app = typer.Typer()
app.command()(sequences)


# noinspection PyUnusedLocal
def test_multi_chain():

    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    result = run_and_report(app, ["-o"], input=STREAM)

    EXPECTED = """\
        ------------- A.seq -------------
        his 3
        met
        ------------- B.seq -------------
        arg 5
        gln
        ------------- C.seq -------------
        pro 7
        ---------------------------------
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain_template():

    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    result = run_and_report(app, ["-o", "--template", "test_%s.seq"], input=STREAM)

    EXPECTED = """\
        ------------- test_A.seq -------------
        his 3
        met
        ------------- test_B.seq -------------
        arg 5
        gln
        ------------- test_C.seq -------------
        pro 7
        --------------------------------------
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain_file_names_one_filename():

    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    result = run_and_report(app, ["-o", "--file-names", "wibble_a.seq"], input=STREAM)

    EXPECTED = """\
        ------------- wibble_a.seq -------------
        his 3
        met
        -------------    B.seq     -------------
        arg 5
        gln
        -------------    C.seq     -------------
        pro 7
        ----------------------------------------
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain_file_names_too_many():

    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    result = run_and_report(
        app,
        ["-o", "--file-names", "wibble_a.seq,wibble_b.seq,wibble_c.seq,wibble_d.seq"],
        input=STREAM,
        expected_exit_code=1,
    )

    assert "there are more file names than chains" in result.stdout


def test_multi_chain_bad_selector():

    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    result = run_and_report(app, ["-o", "AAAA"], input=STREAM, expected_exit_code=1)

    assert "the chain code" in result.stdout
    assert "not in the molecular systems chain codes" in result.stdout
    assert "AAAA" in result.stdout


def test_multi_chain_file_names_one_filename_disk(tmp_path):
    STREAM = open(path_in_test_data(__file__, "multi_chain.nef")).read()

    os.chdir(tmp_path)

    run_and_report(app, ["--file-names", "wibble_a.seq"], input=STREAM)

    EXPECTED_FILES = "wibble_a.seq", "B.seq", "C.seq"

    for file_name in EXPECTED_FILES:
        assert (tmp_path / file_name).is_file()

    EXPECTED = {
        "wibble_a.seq": """\
            his 3
            met
        """,
        "B.seq": """\
            arg 5
            gln
        """,
        "C.seq": """\
            pro 7
        """,
    }

    for file_name in EXPECTED_FILES:
        with open(file_name) as fp:
            text = fp.read()

            assert text == dedent(EXPECTED[file_name])



from pathlib import Path
from textwrap import dedent, indent
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry

from nef_pipelines.lib.nef_frames_lib import (
    EXPERIMENT_CLASSIFICATION,
    EXPERIMENT_TYPE,
    SHIFT_LIST_FRAME_CATEGORY,
)
from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.spectra_lib import (
    EXPERIMENT_CLASSIFICATION_TO_SYNONYM,
    EXPERIMENT_INFO,
    SPECTRUM_TYPE_TO_CLASSIFICATION,
    TRIPLE_SPECTRA,
    ExperimentType,
    PeakInfo,
)
from nef_pipelines.lib.structures import NewPeak, ShiftData
from nef_pipelines.lib.util import (
    FOUR_SPACES,
    STDIN,
    exit_error,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
)
from nef_pipelines.tools.shifts import shifts_app

FAKE_SPECTROMETER_FREQUENCY_600 = 600.12345678
SPECTRUM_NAME_TEMPLATE = "synthetic_{spectrum}"

SHIFT_FRAMES_HELP = """\
    the names of the shift frames to use, this can be a comma separated list of name or the option can be called
    called multiple times. Wild cards are allowed. If no match is found wild cards are checked as well unless
    --exact is set
"""

SPECTRA_HELP = f"""
    the names of the spectra to create, this can be a comma separated list of options or
    called multiple times. Possible values are: {', '.join(ExperimentType)}
"""
NAME_TEMPLATE_HELP = """
    the name template for the new peak lists the placeholder {spectrum} will get replaced with the spectrum type
"""


@shifts_app.command()
def make_peaks(
    shift_frame_selectors: List[str] = typer.Option(
        ["default"], help=SHIFT_FRAMES_HELP
    ),
    exact: bool = typer.Option(
        False, help="if set frames are selected by exact matches"
    ),
    spectra: List[str] = typer.Option([ExperimentType.N_HSQC], help=SPECTRA_HELP),
    name_template: str = typer.Option(SPECTRUM_NAME_TEMPLATE, help=NAME_TEMPLATE_HELP),
    spectrometer_frequency: float = typer.Option(FAKE_SPECTROMETER_FREQUENCY_600),
    input: Path = typer.Option(
        STDIN,
        "--in",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
):
    """-  make a set of peaks for an hsqc or triple resonance spectrum from a list of shiftsdbxcli put rh"""

    shift_frames = parse_comma_separated_options(shift_frame_selectors)

    spectra = set(parse_comma_separated_options(spectra))

    spectra = _update_spectra_with_groups(spectra)

    # TODO add a check that we have some shift frames

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    entry = pipe(
        entry, shift_frames, exact, spectra, name_template, spectrometer_frequency
    )

    print(entry)


def _update_spectra_with_groups(spectra):
    if ExperimentType.TRIPLE in spectra:
        spectra.update(TRIPLE_SPECTRA)
        spectra.remove(ExperimentType.TRIPLE)

    return spectra


def _exit_bad_spectrum_type(spectrum, known_spectra):
    msg = """
        the spectrum type {spectrum} is not know, available spectrum types are

        {spectrum_types}
    """
    msg = dedent(msg)

    spectrum_types = strings_to_tabulated_terminal_sensitive(known_spectra)
    spectrum_types = indent(spectrum_types, FOUR_SPACES)

    exit_error(f(msg))


def pipe(
    entry: Entry,
    shift_frame_selectors: List[str],
    exact_frame_selectors: bool,
    spectra: List[ExperimentType],
    name_template_string: str,
    spectrometer_frequency: float = FAKE_SPECTROMETER_FREQUENCY_600,
) -> Entry:

    all_shift_list_frames = entry.get_saveframes_by_category(SHIFT_LIST_FRAME_CATEGORY)

    selected_shift_list_frames = select_frames_by_name(
        all_shift_list_frames, shift_frame_selectors, exact=exact_frame_selectors
    )

    shifts = nef_frames_to_shifts(selected_shift_list_frames)

    for spectrum in spectra:

        spectrum = spectrum.lower()

        frame_name = f(name_template_string)

        if spectrum not in EXPERIMENT_INFO:
            _exit_bad_spectrum_type(spectrum, list(EXPERIMENT_INFO.keys()))
        spectrum_info = EXPERIMENT_INFO[spectrum]

        peaks = make_spectrum(shifts, spectrum_info)

        dimensions = [
            {"axis_code": dimension} for dimension in spectrum_info.dimensions
        ]

        # TODO add info about frames for errors
        spectrum_classification = SPECTRUM_TYPE_TO_CLASSIFICATION[spectrum]
        extra_tags = {
            EXPERIMENT_CLASSIFICATION: spectrum_classification,
            EXPERIMENT_TYPE: EXPERIMENT_CLASSIFICATION_TO_SYNONYM[
                spectrum_classification
            ],
        }
        frame = peaks_to_frame(
            peaks, dimensions, spectrometer_frequency, frame_name, extra_tags=extra_tags
        )

        entry.add_saveframe(frame)

    return entry


def make_spectrum(
    shifts: List[ShiftData], info: PeakInfo, height: int = 1_000_000
) -> List[NewPeak]:

    # note this is not very sophisticated and doesn't cover side-chains... we ought to port the ccpn code
    relevant_shifts = []

    for atom_name in info.atoms:
        for shift in shifts:
            shift_atom_name = shift.atom.atom_name.upper()
            shift_atom_name = shift_atom_name.strip("1")
            shift_atom_name = shift_atom_name.strip("-")

            if shift_atom_name in atom_name:
                relevant_shifts.append(shift)

    shifts_by_residue = {}

    for shift in relevant_shifts:
        residue = str(shift.atom.residue.sequence_code).split("-")[0]
        shifts_by_residue.setdefault(residue, []).append(shift)

    peaks = []

    for residue_shifts in shifts_by_residue.values():

        for atom_set, sign in zip(info.atom_sets, info.atom_set_signs):

            atom_set_shifts = []
            for number_sequence_code_fields_required, atom_name_required in atom_set:
                for shift in residue_shifts:

                    # these are errors for the spectra we are interested in currently
                    # maybe add a warning?
                    if "+" in str(shift.atom.residue.sequence_code):
                        continue

                    number_sequence_code_fields = len(
                        str(shift.atom.residue.sequence_code).split("-")
                    )
                    atom_name = shift.atom.atom_name

                    number_fields_ok = (
                        number_sequence_code_fields
                        == number_sequence_code_fields_required
                    )
                    atom_name_ok = atom_name_required in atom_name

                    if number_fields_ok and atom_name_ok:
                        atom_set_shifts.append(shift)

            if len(atom_set_shifts) == len(info.atoms):
                new_peak = NewPeak(
                    atom_set_shifts, height=height * sign, volume=height * sign
                )
                peaks.append(new_peak)

    return peaks



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.spectra_lib import (
    EXPERIMENT_CLASSIFICATION_TO_SYNONYM,
    SPECTRUM_TYPE_TO_CLASSIFICATION,
    ExperimentType,
)
from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.shifts.make_peaks import make_peaks

runner = CliRunner()
app = typer.Typer()
app.command()(make_peaks)

EXPECTED = """ # noqa: E501
save_nef_nmr_spectrum_synthetic_n_hsqc
   _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
   _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_synthetic_n_hsqc
   _nef_nmr_spectrum.num_dimensions             2
   _nef_nmr_spectrum.chemical_shift_list        .
   _nef_nmr_spectrum.experiment_classification  H[N]
   _nef_nmr_spectrum.experiment_type            '15N HSQC/HMQC'

   loop_
      _nef_spectrum_dimension.dimension_id
      _nef_spectrum_dimension.axis_unit
      _nef_spectrum_dimension.axis_code
      _nef_spectrum_dimension.spectrometer_frequency
      _nef_spectrum_dimension.spectral_width
      _nef_spectrum_dimension.value_first_point
      _nef_spectrum_dimension.folding
      _nef_spectrum_dimension.absolute_peak_positions
      _nef_spectrum_dimension.is_acquisition

     1   ppm   N15   60.833    24.487   125.828   none   true   false
     2   ppm   H1    600.123   0.606    8.535     none   true   true

   stop_

   loop_
      _nef_spectrum_dimension_transfer.dimension_1
      _nef_spectrum_dimension_transfer.dimension_2
      _nef_spectrum_dimension_transfer.transfer_type
      _nef_spectrum_dimension_transfer.is_indirect

     1   2   onebond   false

   stop_

   loop_
      _nef_peak.index
      _nef_peak.peak_id
      _nef_peak.chain_code_1
      _nef_peak.sequence_code_1
      _nef_peak.residue_name_1
      _nef_peak.atom_name_1
      _nef_peak.chain_code_2
      _nef_peak.sequence_code_2
      _nef_peak.residue_name_2
      _nef_peak.atom_name_2
      _nef_peak.position_1
      _nef_peak.position_uncertainty_1
      _nef_peak.position_2
      _nef_peak.position_uncertainty_2
      _nef_peak.height
      _nef_peak.height_uncertainty
      _nef_peak.volume
      _nef_peak.volume_uncertainty

     0   0   '#18'   @19   .     N   '#18'   @19   .     H   123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     1   1   @-      @9    .     N   @-      @9    .     H   124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     2   2   A       10    Ala   N   A       10    Ala   H   102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .

   stop_

save_
""".replace(
    NOQA_E501, ""
)


# noinspection PyUnusedLocal
def test_ubi_short():

    input = open(path_in_test_data(__file__, "ubiquitin_short.nef")).read()
    result = run_and_report(app, [], input=input)

    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_n_hsqc")
    assert_lines_match(EXPECTED, new_peaks_frame)


def test_ubi_short_different_frame_name():
    input = open(path_in_test_data(__file__, "ubiquitin_short.nef")).read()
    result = run_and_report(app, ["--name-template", "wibble_{spectrum}"], input=input)

    isolate_frame(result.stdout, "nef_nmr_spectrum_wibble_n_hsqc")


def test_ubi_short_different_spectrometer_frequency():
    input = open(path_in_test_data(__file__, "ubiquitin_short.nef")).read()
    result = run_and_report(app, ["--spectrometer-frequency", "800"], input=input)

    EXPECTED_800 = EXPECTED.replace("60.833", "81.095")
    EXPECTED_800 = EXPECTED_800.replace("600.123", "800.000")
    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_n_hsqc")
    assert_lines_match(EXPECTED_800, new_peaks_frame)


EXPECTED_HNCA = """ # noqa: E501
save_nef_nmr_spectrum_synthetic_hnca
   _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
   _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_synthetic_hnca
   _nef_nmr_spectrum.num_dimensions             3
   _nef_nmr_spectrum.chemical_shift_list        .
   _nef_nmr_spectrum.experiment_classification  H[N[CA]]
   _nef_nmr_spectrum.experiment_type            HNCA

   loop_
      _nef_spectrum_dimension.dimension_id
      _nef_spectrum_dimension.axis_unit
      _nef_spectrum_dimension.axis_code
      _nef_spectrum_dimension.spectrometer_frequency
      _nef_spectrum_dimension.spectral_width
      _nef_spectrum_dimension.value_first_point
      _nef_spectrum_dimension.folding
      _nef_spectrum_dimension.absolute_peak_positions
      _nef_spectrum_dimension.is_acquisition

     1   ppm   C13   150.933   9.523    61.581    none   true   false
     2   ppm   N15   60.833    24.487   125.828   none   true   false
     3   ppm   H1    600.123   0.606    8.535     none   true   true

   stop_

   loop_
      _nef_spectrum_dimension_transfer.dimension_1
      _nef_spectrum_dimension_transfer.dimension_2
      _nef_spectrum_dimension_transfer.dimension_3
      _nef_spectrum_dimension_transfer.transfer_type
      _nef_spectrum_dimension_transfer.is_indirect

     .   2   3   onebond   false
     .   2   3   onebond   false

   stop_

   loop_
      _nef_peak.index
      _nef_peak.peak_id
      _nef_peak.chain_code_1
      _nef_peak.sequence_code_1
      _nef_peak.residue_name_1
      _nef_peak.atom_name_1
      _nef_peak.chain_code_2
      _nef_peak.sequence_code_2
      _nef_peak.residue_name_2
      _nef_peak.atom_name_2
      _nef_peak.chain_code_3
      _nef_peak.sequence_code_3
      _nef_peak.residue_name_3
      _nef_peak.atom_name_3
      _nef_peak.position_1
      _nef_peak.position_uncertainty_1
      _nef_peak.position_2
      _nef_peak.position_uncertainty_2
      _nef_peak.position_3
      _nef_peak.position_uncertainty_3
      _nef_peak.height
      _nef_peak.height_uncertainty
      _nef_peak.volume
      _nef_peak.volume_uncertainty

     0   0   '#18'   @19     .     CA   '#18'   @19   .     N   '#18'   @19   .     H   55.19626104   .              123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     1   1   '#18'   @19-1   .     CA   '#18'   @19   .     N   '#18'   @19   .     H   56.6890552    0.0499432144   123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     2   2   @-      @9      .     CA   @-      @9    .     N   @-      @9    .     H   57.45044906   .              124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     3   3   @-      @9-1    .     CA   @-      @9    .     N   @-      @9    .     H   61.14832291   0.0266869088   124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     4   4   A       10-1    Ala   CA   A       10    Ala   N   A       10    Ala   H   52.49098528   0              102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .

   stop_

save_

""".replace(
    NOQA_E501, ""
)


def test_ubi_short_triple():

    input = open(path_in_test_data(__file__, "ubiquitin_short.nef")).read()
    result = run_and_report(app, ["--spectra", "HNCA"], input=input)

    print(result.stdout)
    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_hnca")
    assert_lines_match(EXPECTED_HNCA, new_peaks_frame)


EXPECTED_EXPERIMENT_CLASSIFICATIONS_AND_SYNONYMS = {
    ExperimentType.N_HSQC: ("H[N]", "15N HSQC/HMQC"),
    ExperimentType.C_HSQC: ("H[C]", "13C HSQC/HMQC"),
    ExperimentType.HNCA: ("H[N[CA]]", "HNCA"),
    ExperimentType.HNcoCA: ("H[N[co[CA]]]", "H-detected HNcoCA"),
    ExperimentType.HNCACB: ("H[N[CA[CB]]]", "H-detected HNCACB"),
    ExperimentType.HNcoCACB: ("H[N[co[CA[CB]]]]", "H-detected HNcoCACB"),
    ExperimentType.CBCAcoNH: ("h{CA|Cca}coNH", "hbCB/haCAcoNNH"),
    ExperimentType.HNCO: ("H[N[CO]]", "HNCO"),
    ExperimentType.HNcaCO: ("H[N[ca[CO]]]", "H-detected HNcaCO"),
}


def test_experiment_classifications_and_synonyms():
    for experiment_type in ExperimentType:
        if experiment_type == ExperimentType.TRIPLE:
            continue
        classification = SPECTRUM_TYPE_TO_CLASSIFICATION[experiment_type]
        synonym = EXPERIMENT_CLASSIFICATION_TO_SYNONYM[classification]

        assert EXPECTED_EXPERIMENT_CLASSIFICATIONS_AND_SYNONYMS[experiment_type] == (
            classification,
            synonym,
        )



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrpipe.importers.peaks import peaks

HEADER = open(path_in_test_data(__file__, "test_header_entry.txt")).read()

PEAKS_NMRPIPE = "nef_nmr_spectrum_nmrpipe"
METADATA_NMRPIPE = "nef_nmr_meta_data"


app = typer.Typer()
app.command()(peaks)


# noinspection PyUnusedLocal
def test_peaks():
    EXPECTED = open(
        path_in_test_data(__file__, "gb3_assigned_trunc_expected.tab")
    ).read()

    path = path_in_test_data(__file__, "gb3_assigned_trunc.tab")
    result = run_and_report(app, [path], input=HEADER)

    assert result.exit_code == 0
    peaks_result = isolate_frame(result.stdout, "%s" % PEAKS_NMRPIPE)

    assert_lines_match(EXPECTED, peaks_result)



from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Tuple, Union

from strenum import LowercaseStrEnum

from nef_pipelines.lib.isotope_lib import Isotope


class NMRMassClass(Enum):
    LOW = auto()
    HIGH = auto()


class ExperimentType(LowercaseStrEnum):
    N_HSQC = auto()
    C_HSQC = auto()

    HNCA = auto()
    HNcoCA = auto()

    HNCACB = auto()
    HNcoCACB = auto()
    CBCAcoNH = auto()

    HNCO = auto()
    HNcaCO = auto()

    TRIPLE = auto()


SPECTRUM_TYPE_TO_CLASSIFICATION = {
    ExperimentType.N_HSQC: "H[N]",
    ExperimentType.C_HSQC: "H[C]",
    ExperimentType.HNCA: "H[N[CA]]",
    ExperimentType.HNcoCA: "H[N[co[CA]]]",
    ExperimentType.HNCACB: "H[N[CA[CB]]]",  # even though you might think of it as H[N[{CA|CB}]]'
    # or H[N[{CA|ca[Cali]}]]]
    # or even H[N[{CA|ca[CB]}]]
    # or even H[N[{CA|ca[C]}]]  ?
    ExperimentType.HNcoCACB: "H[N[co[CA[CB]]]]",  # even though you might think of it related to ones above
    # or 'H[N[co[{CA|ca[C]}]]]
    ExperimentType.CBCAcoNH: "h{CA|Cca}coNH",  # and why is this not h{CA|CBca}coNH
    ExperimentType.HNCO: "H[N[CO]]",
    ExperimentType.HNcaCO: "H[N[ca[CO]]]",
}

TRIPLE_SPECTRA = {
    ExperimentType.N_HSQC,
    ExperimentType.HNCA,
    ExperimentType.HNcoCA,
    ExperimentType.HNCACB,
    ExperimentType.HNcoCACB,
    ExperimentType.HNCO,
    ExperimentType.HNcaCO,
}


@dataclass
class PeakInfo:
    atoms: Tuple[Union[str, Tuple[str, ...]], ...]
    dimensions: Tuple[Isotope, ...]
    atom_sets: Tuple[Tuple[Tuple[int, str], ...], ...]
    atom_set_signs: Tuple[int, ...]
    essential_atom_sets: Tuple[int, ...]
    atom_sets_by_mass: Dict[NMRMassClass, Tuple[int, ...]]
    atom_set_names: Tuple[str, ...]
    partners: Tuple[ExperimentType, ...]


@dataclass
class AtomInfo:
    n_field: int
    atom_type: str


# TODO much of this is ugly!
EXPERIMENT_INFO = {
    ExperimentType.N_HSQC: PeakInfo(
        atoms=("N", "H"),
        dimensions=(Isotope.N15, Isotope.H1),
        atom_sets=(((1, "N"), (1, "H")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("HN",),
        partners=(),
    ),
    # SpectraTypes.C_HSQC: ShiftInfo(
    #     atoms={'H', 'C'},
    #     dimensions= (Isotope.H1, Isotope.N15),
    #     atom_sets=[[(1, 'H'), (1, 'C')],],
    #     atom_set_signs=[1]
    # ),
    ExperimentType.HNCA: PeakInfo(
        atoms=(("CA",), "N", "H"),
        dimensions=(Isotope.C13, Isotope.N15, Isotope.H1),
        atom_sets=(
            ((1, "CA"), (1, "N"), (1, "H")),
            ((2, "CA"), (1, "N"), (1, "H")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA", "CA-1"),
        partners=(ExperimentType.HNcoCA,),
    ),
    ExperimentType.HNcoCA: PeakInfo(
        atoms=("H", "N", ("CA",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(((1, "H"), (1, "N"), (2, "CA")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("CA-1",),
        partners=(ExperimentType.HNCA,),
    ),
    ExperimentType.HNCACB: PeakInfo(
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((1, "H"), (1, "N"), (1, "CA")),
            ((1, "H"), (1, "N"), (2, "CA")),
            ((1, "H"), (1, "N"), (1, "CB")),
            ((1, "H"), (1, "N"), (2, "CB")),
        ),
        atom_set_signs=(1, 1, -1, -1),
        essential_atom_sets=(0, 2),
        atom_sets_by_mass={
            NMRMassClass.LOW: (0, 1, 2, 3),
            NMRMassClass.HIGH: (0, 1, 2),
        },
        atom_set_names=("CA", "CA-1", "CB", "CB-1"),
        partners=(ExperimentType.HNcoCACB, ExperimentType.CBCAcoNH),
    ),
    ExperimentType.HNcoCACB: PeakInfo(
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((1, "H"), (1, "N"), (2, "CA")),
            ((1, "H"), (1, "N"), (2, "CB")),
        ),
        atom_set_signs=(1, -1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA-1", "CA-1"),
        partners=(ExperimentType.HNCACB,),
    ),
    ExperimentType.CBCAcoNH: PeakInfo(
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((1, "H"), (1, "N"), (2, "CA")),
            ((1, "H"), (1, "N"), (2, "CB")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA-1", "CB-1"),
        partners=(ExperimentType.HNCACB,),
    ),
    ExperimentType.HNCO: PeakInfo(
        atoms=("H", "N", ("C",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(((1, "H"), (1, "N"), (2, "C")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("C-1",),
        partners=(ExperimentType.HNcaCO,),
    ),
    ExperimentType.HNcaCO: PeakInfo(
        atoms=("H", "N", ("C",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((1, "H"), (1, "N"), (1, "C")),
            ((1, "H"), (1, "N"), (2, "C")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (1,)},
        atom_set_names=("C", "C-1"),
        partners=(ExperimentType.HNCO,),
    ),
}

# a subset of the synonyms produced by lib/experiments/prototypes from the ccp v2 experiment prototypes
# just fro tipe reoenance spectra and hsqcs
EXPERIMENT_SYNONYM_TO_CLASSIFICATION = {
    "15N HSQC/HMQC": "H[N]",
    "13C HSQC/HMQC": "H[C]",
    "H-detected HNcoCA": "H[N[co[CA]]]",
    "HNCA": "H[N[CA]]",
    "H-detected HNCACB": "H[N[CA[CB]]]",
    "H-detected HNcoCACB": "H[N[co[CA[CB]]]]",
    "hbCB/haCAcoNNH": "h{CA|Cca}coNH",
    "HNCO": "H[N[CO]]",
    "H-detected HNcaCO": "H[N[ca[CO]]]",
}

EXPERIMENT_CLASSIFICATION_TO_SYNONYM = {
    synonym: spectrum_type
    for spectrum_type, synonym in EXPERIMENT_SYNONYM_TO_CLASSIFICATION.items()
}
